#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Handles all requests related to resource locks.
"""

from oslo_log import log as logging

from manila.common import constants
from manila.db import base
from manila import exception
from manila import policy

LOG = logging.getLogger(__name__)


class API(base.Base):
    """API for handling resource locks."""

    resource_get = {
        "share": "share_get",
        "access_rule": "share_access_get_with_context"
    }
    resource_lock_disallowed_statuses = {
        "share": constants.DISALLOWED_STATUS_WHEN_LOCKING_SHARES,
        "access_rule": constants.DISALLOWED_STATUS_WHEN_LOCKING_ACCESS_RULES
    }

    def _get_lock_context(self, context):
        if context.is_service:
            lock_context = 'service'
        elif context.is_admin:
            lock_context = 'admin'
        else:
            lock_context = 'user'
        return {
            'lock_context': lock_context,
            'user_id': context.user_id,
            'project_id': context.project_id,
        }

    def _check_allow_lock_manipulation(self, context, resource_lock):
        """Lock owners may not manipulate a lock if lock_context disallows

        The logic enforced by this method is that user created locks can be
        manipulated by all roles, service created locks can be manipulated
        by service and admin roles, while admin created locks can only be
        manipulated by admin role:

        +------------+------------+--------------+---------+
        | Requester  | Lock Owner | Lock Context | Allowed |
        +------------+------------+--------------+---------+
        | user       | user       | user         | yes     |
        | user       | user       | service      | no      |
        | user       | admin      | admin        | no      |
        | admin      | user       | user         | yes     |
        | admin      | user       | service      | yes     |
        | admin      | admin      | admin        | yes     |
        | service    | user       | user         | yes     |
        | service    | user       | service      | yes     |
        | service    | admin      | admin        | no      |
        +------------+------------+--------------+---------+
        """
        locked_by = resource_lock['lock_context']
        update_requested_by = self._get_lock_context(context)['lock_context']
        if ((locked_by == 'admin' and update_requested_by != 'admin')
                or (locked_by == 'service' and update_requested_by == 'user')):
            raise exception.NotAuthorized("Resource lock cannot be "
                                          "manipulated by user. Please "
                                          "contact the administrator.")

    def access_is_restricted(self, context, resource_lock):
        """Ensure the requester doesn't have visibility restrictions

        Call the check allow lock manipulation method as a first validation.
        In case it fails, the requester should not have the access rules
        fields entirely visible. In case it passes and the access visibility
        is restricted, the users will have visibility of all fields only if
        they have originally created the lock.
        """
        try:
            self._check_allow_lock_manipulation(context, resource_lock)
        except exception.NotAuthorized:
            return True

        try:
            policy.check_policy(
                context, 'resource_lock', 'bypass_locked_show_action',
                resource_lock)
        except exception.NotAuthorized:
            return True

        return False

    def get(self, context, lock_id):
        """Return resource lock with the specified id."""
        return self.db.resource_lock_get(context, lock_id)

    def get_all(self, context, search_opts=None, limit=None,
                offset=None, sort_key="created_at", sort_dir="desc",
                show_count=False):
        """Return resource locks for the given context."""
        LOG.debug("Searching for locks by: %s", search_opts)

        search_opts = search_opts or {}
        if 'all_projects' in search_opts:
            allow_all_projects = policy.check_policy(
                context,
                'resource_lock',
                'get_all_projects',
                do_raise=False
            )
            if not allow_all_projects:
                LOG.warning("User %s not allowed to query locks "
                            "across all projects.", context.user_id)
                search_opts.pop('all_projects')
                search_opts.pop('project_id', None)

        locks, count = self.db.resource_lock_get_all(
            context,
            filters=search_opts,
            limit=limit, offset=offset,
            sort_key=sort_key,
            sort_dir=sort_dir,
            show_count=show_count,
        )

        return locks, count

    def create(self, context, resource_id=None, resource_type=None,
               resource_action=None, lock_reason=None, resource=None):
        """Create a resource lock with the specified information."""
        get_res_method = getattr(self.db, self.resource_get[resource_type])
        if resource_action == constants.RESOURCE_ACTION_SHOW:
            # We can't allow visibility locks to be placed more than once,
            # otherwise the resource might become visible to someone else.
            visibility_locks, __ = self.db.resource_lock_get_all(
                context.elevated(),
                filters={'resource_id': resource_id,
                         'resource_action': resource_action,
                         'all_projects': True})
            if visibility_locks:
                raise exception.ResourceVisibilityLockExists(
                    resource_id=resource_id)
        if resource is None:
            resource = get_res_method(context, resource_id)
        policy.check_policy(context, 'resource_lock', 'create', resource)
        self._check_resource_state_for_locking(
            resource_action, resource, resource_type=resource_type)
        lock_context_data = self._get_lock_context(context)
        resource_lock = lock_context_data.copy()
        resource_lock.update({
            'resource_id': resource_id,
            'resource_action': resource_action,
            'lock_reason': lock_reason,
            'resource_type': resource_type
        })
        return self.db.resource_lock_create(context, resource_lock)

    def _check_resource_state_for_locking(self, resource_action, resource,
                                          resource_type='share'):
        """Check if resource is in a "disallowed" state for locking.

        For example, deletion lock on a "deleting" resource would be futile.
        """
        resource_state = resource.get('status', resource.get('state', ''))
        disallowed_statuses = ()
        if resource_action == 'delete':
            disallowed_statuses = (
                self.resource_lock_disallowed_statuses[resource_type])
        if resource_state in disallowed_statuses:
            msg = "Resource status not suitable for locking"
            raise exception.InvalidInput(reason=msg)
        if resource_type == constants.SHARE_RESOURCE_TYPE:
            resource_is_soft_deleted = resource.get('is_soft_deleted', False)
            if resource_is_soft_deleted:
                msg = (
                    "Resource cannot be locked since it has been soft deleted."
                )
                raise exception.InvalidInput(reason=msg)

    def update(self, context, resource_lock, updates):
        """Update a resource lock with the specified information."""
        lock_id = resource_lock['id']
        policy.check_policy(context, 'resource_lock', 'update', resource_lock)
        self._check_allow_lock_manipulation(context, resource_lock)
        if 'resource_action' in updates:
            # A resource can have only one visibility lock
            if (updates['resource_action'] == constants.RESOURCE_ACTION_SHOW
                    and resource_lock['resource_action'] !=
                    constants.RESOURCE_ACTION_SHOW):
                filters = {
                    "resource_id": resource_lock['resource_id'],
                    "resource_action": constants.RESOURCE_ACTION_SHOW
                }
                visibility_locks = self.get_all(
                    context.elevated(), search_opts=filters)
                if visibility_locks:
                    msg = "The resource already has a visibility lock."
                    raise exception.InvalidInput(reason=msg)
            get_res_method = getattr(
                self.db,
                self.resource_get[resource_lock['resource_type']],
            )
            resource = get_res_method(context, resource_lock['resource_id'])
            self._check_resource_state_for_locking(
                updates['resource_action'], resource)
        return self.db.resource_lock_update(context, lock_id, updates)

    def ensure_context_can_delete_lock(self, context, lock_id):
        """Ensure the requester is able to delete locks."""
        resource_lock = self.db.resource_lock_get(context, lock_id)
        policy.check_policy(context, 'resource_lock', 'delete', resource_lock)
        self._check_allow_lock_manipulation(context, resource_lock)

    def delete(self, context, lock_id):
        """Delete resource lock with the specified id."""
        self.ensure_context_can_delete_lock(context, lock_id)
        self.db.resource_lock_delete(context, lock_id)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Common field types for validating API responses."""

links = {
    'type': 'array',
    'items': {
        'type': 'object',
        'properties': {
            'rel': {
                'type': 'string',
                'enum': ['self', 'bookmark'],
            },
            'href': {
                'type': 'string',
                'format': 'uri',
            },
        },
        'required': ['rel', 'href'],
        'additionalProperties': False,
    },
}

collection_links = {
    'type': 'array',
    'items': {
        'type': 'object',
        'properties': {
            'rel': {
                'const': 'next',
            },
            'href': {
                'type': 'string',
                'format': 'uri',
            },
        },
        'required': ['rel', 'href'],
        'additionalProperties': False,
    },
    # there should be one and only one link object
    'minItems': 1,
    'maxItems': 1,
}



# Copyright 2024 VAST Data Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
VAST's Share Driver


Configuration:


[DEFAULT]
enabled_share_backends = vast

[vast]
share_driver = manila.share.drivers.vastdata.driver.VASTShareDriver
share_backend_name = vast
snapshot_support = true
driver_handles_share_servers = false
vast_mgmt_host = v11
vast_vippool_name = vippool-1
vast_root_export = manila
vast_mgmt_user = admin
vast_mgmt_password = 123456
"""

import collections

import netaddr
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import units

from manila.common import constants
from manila import exception
from manila.i18n import _
from manila.share import driver
from manila.share.drivers.vastdata import driver_util
import manila.share.drivers.vastdata.rest as vast_rest


LOG = logging.getLogger(__name__)

OPTS = [
    cfg.HostAddressOpt(
        "vast_mgmt_host",
        help="Hostname or IP address VAST storage system management VIP.",
    ),
    cfg.PortOpt(
        "vast_mgmt_port",
        help="Port for VAST management",
        default=443
    ),
    cfg.StrOpt(
        "vast_vippool_name",
        help="Name of Virtual IP pool"
    ),
    cfg.StrOpt(
        "vast_root_export",
        default="manila",
        help="Base path for shares"
    ),
    cfg.StrOpt(
        "vast_mgmt_user",
        help="Username for VAST management"
    ),
    cfg.StrOpt(
        "vast_mgmt_password",
        help="Password for VAST management",
        secret=True
    ),
]

CONF = cfg.CONF
CONF.register_opts(OPTS)

MANILA_TO_VAST_ACCESS_LEVEL = {
    constants.ACCESS_LEVEL_RW: "nfs_read_write",
    constants.ACCESS_LEVEL_RO: "nfs_read_only",
}


@driver_util.decorate_methods_with(
    driver_util.verbose_driver_trace
)
class VASTShareDriver(driver.ShareDriver):
    """Driver for the VastData Filesystem."""

    VERSION = "1.0"  # driver version

    def __init__(self, *args, **kwargs):
        super().__init__(False, *args, config_opts=[OPTS], **kwargs)

    def do_setup(self, context):
        """Driver initialization"""
        backend_name = self.configuration.safe_get("share_backend_name")
        root_export = self.configuration.vast_root_export
        vip_pool_name = self.configuration.safe_get("vast_vippool_name")
        if not vip_pool_name:
            raise exception.VastDriverException(
                reason="vast_vippool_name must be set"
            )
        self._backend_name = backend_name or self.__class__.__name__
        self._vippool_name = vip_pool_name
        self._root_export = "/" + root_export.strip("/")

        username = self.configuration.safe_get("vast_mgmt_user")
        password = self.configuration.safe_get("vast_mgmt_password")
        host = self.configuration.safe_get("vast_mgmt_host")
        port = self.configuration.safe_get("vast_mgmt_port")
        if not all((username, password, port)):
            raise exception.VastDriverException(
                reason="Not all required parameters are present."
                       " Make sure you specified `vast_mgmt_host`,"
                       " `vast_mgmt_port`, and `vast_mgmt_user` "
                       "in manila.conf."
            )
        if port:
            host = f"{host}:{port}"
        self.rest = vast_rest.RestApi(
            host, username, password, False, self.VERSION
        )
        LOG.debug("VAST Data driver setup is complete.")

    def _update_share_stats(self, data=None):
        """Retrieve stats info from share group."""
        metrics_list = [
            "Capacity,drr",
            "Capacity,logical_space",
            "Capacity,logical_space_in_use",
            "Capacity,physical_space",
            "Capacity,physical_space_in_use",
        ]
        metrics = self.rest.capacity_metrics.get(metrics_list)
        data = dict(
            share_backend_name=self._backend_name,
            vendor_name="VAST STORAGE",
            driver_version=self.VERSION,
            storage_protocol="NFS",
            data_reduction=metrics.drr,
            total_capacity_gb=float(metrics.logical_space) / units.Gi,
            free_capacity_gb=float(
                metrics.logical_space - metrics.logical_space_in_use
            )
            / units.Gi,
            provisioned_capacity_gb=float(
                metrics.logical_space_in_use) / units.Gi,
            snapshot_support=True,
            create_share_from_snapshot_support=False,
            mount_snapshot_support=False,
            revert_to_snapshot_support=False,
        )

        super()._update_share_stats(data)

    def _to_volume_path(self, share_id, root=None):
        if not root:
            root = self._root_export
        return f"{root}/manila-{share_id}"

    def create_share(self, context, share, share_server=None):
        return self._ensure_share(share)[0]

    def delete_share(self, context, share, share_server=None):
        """Called to delete a share"""
        share_id = share["id"]
        src = self._to_volume_path(share_id)
        LOG.debug(f"Deleting '{src}'.")
        self.rest.folders.delete(path=src)
        self.rest.views.delete(name=share_id)
        self.rest.quotas.delete(name=share_id)
        self.rest.view_policies.delete(name=share_id)

    def update_access(
        self, context, share, access_rules,
        add_rules, delete_rules, share_server=None
    ):
        """Update access rules for share."""
        rule_state_map = {}

        if not (add_rules or delete_rules):
            add_rules = access_rules

        if share["share_proto"] != "NFS":
            LOG.error("The share protocol flavor is invalid. Please use NFS.")
            return

        valid_add_rules = []
        for rule in (add_rules or []):
            try:
                validate_access_rule(rule)
            except (
                    exception.InvalidShareAccess,
                    exception.InvalidShareAccessLevel,
            ) as exc:
                rule_id = rule["access_id"]
                access_level = rule["access_level"]
                access_to = rule["access_to"]
                LOG.exception(
                    f"Failed to provide {access_level} access to "
                    f"{access_to} (Rule ID: {rule_id}, Reason: {exc}). "
                    "Setting rule to 'error' state."
                )
                rule_state_map[rule['id']] = {'state': 'error'}
            else:
                valid_add_rules.append(rule)

        share_id = share["id"]
        export = self._to_volume_path(share_id)

        LOG.debug(f"Changing access on {share_id}.")
        data = {
            "name": share_id,
            "nfs_no_squash": ["*"],
            "nfs_root_squash": ["*"]
        }
        policy = self.rest.view_policies.one(name=share_id)
        if not policy:
            raise exception.VastDriverException(
                reason=f"Policy not found for share {share_id}."
            )
        if valid_add_rules:
            policy_rules = policy_payload_from_rules(
                rules=valid_add_rules, policy=policy, action="update"
            )
            data.update(policy_rules)
            LOG.debug(f"Changing access on {export}. Rules: {policy_rules}.")
            self.rest.view_policies.update(policy.id, **data)

        if delete_rules:
            policy_rules = policy_payload_from_rules(
                rules=delete_rules, policy=policy, action="deny"
            )
            LOG.debug(f"Changing access on {export}. Rules: {policy_rules}.")
            data.update(policy_rules)
            self.rest.view_policies.update(policy.id, **data)

        return rule_state_map

    def extend_share(self, share, new_size, share_server=None):
        """uses resize_share to extend a share"""
        self._resize_share(share, new_size)

    def shrink_share(self, share, new_size, share_server=None):
        """uses resize_share to shrink a share"""
        self._resize_share(share, new_size)

    def create_snapshot(self, context, snapshot, share_server=None):
        """Is called to create snapshot."""
        path = self._to_volume_path(snapshot["share_instance_id"])
        self.rest.snapshots.create(path=path, name=snapshot["name"])

    def delete_snapshot(self, context, snapshot, share_server=None):
        """Is called to remove share."""
        self.rest.snapshots.delete(name=snapshot["name"])

    def get_network_allocations_number(self):
        return 0

    def ensure_shares(self, context, shares):
        updates = {}
        for share in shares:
            export_locations = self._ensure_share(share)
            updates[share["id"]] = {
                'export_locations': export_locations
            }
        return updates

    def get_backend_info(self, context):
        backend_info = {
            "vast_vippool_name": self.configuration.vast_vippool_name,
            "vast_mgmt_host": self.configuration.vast_mgmt_host,
        }
        return backend_info

    def _resize_share(self, share, new_size):
        share_id = share["id"]
        quota = self.rest.quotas.one(name=share_id)
        if not quota:
            raise exception.ShareNotFound(
                reason="Share not found", share_id=share_id
            )
        requested_capacity = new_size * units.Gi
        if requested_capacity < quota.used_effective_capacity:
            raise exception.ShareShrinkingPossibleDataLoss(
                share_id=share['id'])
        self.rest.quotas.update(quota.id, hard_limit=requested_capacity)

    def _ensure_share(self, share):
        share_proto = share["share_proto"]
        if share_proto != "NFS":
            raise exception.InvalidShare(
                reason=_(
                    "Invalid NAS protocol supplied: {}.".format(share_proto)
                )
            )

        vips = self.rest.vip_pools.vips(pool_name=self._vippool_name)

        share_id = share["id"]
        requested_capacity = share["size"] * units.Gi
        path = self._to_volume_path(share_id)
        policy = self.rest.view_policies.ensure(name=share_id)
        quota = self.rest.quotas.ensure(
            name=share_id, path=path,
            create_dir=True, hard_limit=requested_capacity
        )
        if quota.hard_limit != requested_capacity:
            raise exception.VastDriverException(
                reason=f"Share already exists with different capacity"
                f" (requested={requested_capacity}, exists={quota.hard_limit})"
                )
        view = self.rest.views.ensure(
            name=share_id, path=path, policy_id=policy.id
        )
        if view.policy != share_id:
            self.rest.views.update(view.id, policy_id=policy.id)
        return [
            dict(path=f"{vip}:{path}", is_admin_only=False) for vip in vips
        ]


def policy_payload_from_rules(rules, policy, action):
    """Convert list of manila rules

    into vast compatible payload for updating/creating policy.
    """
    hosts = collections.defaultdict(set)
    for rule in rules:
        addr_list = map(
            str, netaddr.IPNetwork(rule["access_to"]).iter_hosts()
        )
        hosts[
            MANILA_TO_VAST_ACCESS_LEVEL[rule["access_level"]]
        ].update(addr_list)

    _default_rules = set()

    # Delete default_vast_policy on each update.
    # There is no sense to keep * in list of allowed/denied hosts
    # as user want to set particular ip/ips only.
    _default_vast_policy = {"*"}
    if action == "update":
        rw = set(policy.nfs_read_write).union(
            hosts.get("nfs_read_write", _default_rules)
        )
        ro = set(policy.nfs_read_only).union(
            hosts.get("nfs_read_only", _default_rules)
        )
    elif action == "deny":
        rw = set(policy.nfs_read_write).difference(
            hosts.get("nfs_read_write", _default_rules)
        )
        ro = set(policy.nfs_read_only).difference(
            hosts.get("nfs_read_only", _default_rules)
        )
    else:
        raise ValueError("Invalid action")

    # When policy created default access is
    # "*" for read-write and read-only operations.
    # After updating any of rules (rw or ro)
    # we need to delete "*" to prevent ambiguous state when
    # resource available for certain ip and for all range of ip addresses.
    if len(rw) > 1:
        rw -= _default_vast_policy

    if len(ro) > 1:
        ro -= _default_vast_policy

    return {"nfs_read_write": list(rw), "nfs_read_only": list(ro)}


def validate_access_rule(access_rule):
    allowed_types = {"ip"}
    allowed_levels = MANILA_TO_VAST_ACCESS_LEVEL.keys()

    access_type = access_rule["access_type"]
    access_level = access_rule["access_level"]
    if access_type not in allowed_types:
        reason = _("Only {} access type allowed.").format(
            ", ".join(tuple([f"'{x}'" for x in allowed_types]))
        )
        raise exception.InvalidShareAccess(reason=reason)
    if access_level not in allowed_levels:
        raise exception.InvalidShareAccessLevel(level=access_level)
    try:
        netaddr.IPNetwork(access_rule["access_to"])
    except (netaddr.core.AddrFormatError, OSError) as exc:
        raise exception.InvalidShareAccess(reason=str(exc))



# Copyright 2024 VAST Data Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import json
import pickle
from unittest import mock

import ddt

from manila.share.drivers.vastdata import driver_util
from manila import test


driver_util.CONF.debug = True


@ddt.ddt
class TestBunch(test.TestCase):
    def setUp(self):
        super(TestBunch, self).setUp()
        self.bunch = driver_util.Bunch(a=1, b=2)

    def test_bunch_getattr(self):
        self.assertEqual(self.bunch.a, 1)

    def test_bunch_setattr(self):
        self.bunch.c = 3
        self.assertEqual(self.bunch.c, 3)

    def test_bunch_delattr(self):
        del self.bunch.a
        self.assertRaises(AttributeError, lambda: self.bunch.a)

    def test_bunch_to_dict(self):
        self.assertEqual(self.bunch.to_dict(), {"a": 1, "b": 2})

    def test_bunch_from_dict(self):
        self.assertEqual(
            driver_util.Bunch.from_dict({"a": 1, "b": 2}), self.bunch
        )

    def test_bunch_to_json(self):
        self.assertEqual(self.bunch.to_json(), json.dumps({"a": 1, "b": 2}))

    def test_bunch_without(self):
        self.assertEqual(self.bunch.without("a"), driver_util.Bunch(b=2))

    def test_bunch_but_with(self):
        self.assertEqual(
            self.bunch.but_with(c=3), driver_util.Bunch(a=1, b=2, c=3)
        )

    def test_bunch_delattr_missing(self):
        self.assertRaises(
            AttributeError,
            lambda: self.bunch.__delattr__("non_existing_attribute")
        )

    def test_bunch_from_json(self):
        json_bunch = json.dumps({"a": 1, "b": 2})
        self.assertEqual(driver_util.Bunch.from_json(json_bunch), self.bunch)

    def test_bunch_render(self):
        self.assertEqual(self.bunch.render(), "a=1, b=2")

    def test_bunch_pickle(self):
        pickled_bunch = pickle.dumps(self.bunch)
        unpickled_bunch = pickle.loads(pickled_bunch)
        self.assertEqual(self.bunch, unpickled_bunch)

    @ddt.data(True, False)
    def test_bunch_copy(self, deep):
        copy_bunch = self.bunch.copy(deep=deep)
        self.assertEqual(copy_bunch, self.bunch)
        self.assertIsNot(copy_bunch, self.bunch)

    def test_name_starts_with_underscore_and_digit(self):
        bunch = driver_util.Bunch()
        bunch["1"] = "value"
        self.assertEqual(bunch._1, "value")

    def test_bunch_recursion(self):
        x = driver_util.Bunch(
            a="a", b="b", d=driver_util.Bunch(x="axe", y="why")
        )
        x.d.x = x
        x.d.y = x.b
        print(x)

    def test_bunch_repr(self):
        self.assertEqual(repr(self.bunch), "Bunch(a=1, b=2)")

    def test_getitem_with_integral_key(self):
        self.bunch["1"] = "value"
        self.assertEqual(self.bunch[1], "value")

    def test_bunch_dir(self):
        self.assertEqual(
            set(i for i in dir(self.bunch) if not i.startswith("_")),
            {
                "a",
                "b",
                "but_with",
                "clear",
                "copy",
                "from_dict",
                "from_json",
                "fromkeys",
                "get",
                "items",
                "keys",
                "pop",
                "popitem",
                "render",
                "setdefault",
                "to_dict",
                "to_json",
                "update",
                "values",
                "without",
            },
        )

    def test_bunch_edge_cases(self):
        # Test edge cases for attribute access, setting, and deletion
        self.bunch["key-with-special-chars_123"] = "value"
        self.assertEqual(self.bunch["key-with-special-chars_123"], "value")
        self.bunch["key-with-special-chars_123"] = None
        self.assertIsNone(self.bunch["key-with-special-chars_123"])
        del self.bunch["key-with-special-chars_123"]
        self.assertRaises(
            KeyError,
            lambda: self.bunch["key-with-special-chars_123"]
        )

    def test_bunch_deep_copy(self):
        nested_bunch = driver_util.Bunch(x=driver_util.Bunch(y=1))
        deep_copy = nested_bunch.copy(deep=True)
        self.assertIsNot(nested_bunch["x"], deep_copy["x"])
        self.assertEqual(nested_bunch["x"]["y"], deep_copy["x"]["y"])

    def test_bunch_serialization(self):
        # Test serialization with nested structures
        nested_bunch = driver_util.Bunch(a=1, b=driver_util.Bunch(c=2))
        self.assertEqual(nested_bunch.to_dict(), {"a": 1, "b": {"c": 2}})
        self.assertEqual(
            nested_bunch.to_json(),
            json.dumps({"a": 1, "b": {"c": 2}})
        )


class TestBunchify(test.TestCase):
    def test_bunchify(self):
        self.assertEqual(
            driver_util.bunchify({"a": 1, "b": 2}, c=3),
            driver_util.Bunch(a=1, b=2, c=3)
        )
        x = driver_util.bunchify(dict(a=[dict(b=5), 9, (1, 2)], c=8))
        self.assertEqual(x.a[0].b, 5)
        self.assertEqual(x.a[1], 9)
        self.assertIsInstance(x.a[2], tuple)
        self.assertEqual(x.c, 8)
        self.assertEqual(x.pop("c"), 8)

    def test_bunchify_edge_cases(self):
        # Test edge cases for bunchify function
        self.assertEqual(driver_util.bunchify({}), driver_util.Bunch())

    def test_bunchify_nested_structures(self):
        # Test bunchify with nested structures
        nested_dict = {"a": [{"b": 1}, 2]}
        self.assertEqual(driver_util.bunchify(nested_dict).a[0].b, 1)


class TestUnbunchify(test.TestCase):
    def test_unbunchify(self):
        self.assertEqual(
            driver_util.unbunchify(driver_util.Bunch(a=1, b=2)),
            {"a": 1, "b": 2}
        )


@ddt.ddt
class TestGenerateIpRange(test.TestCase):

    @ddt.data(
        (
            [["15.0.0.1", "15.0.0.4"], ["10.0.0.27", "10.0.0.30"]],
            [
                "15.0.0.1",
                "15.0.0.2",
                "15.0.0.3",
                "15.0.0.4",
                "10.0.0.27",
                "10.0.0.28",
                "10.0.0.29",
                "10.0.0.30",
            ],
        ),
        (
            [["15.0.0.1", "15.0.0.1"], ["10.0.0.20", "10.0.0.20"]],
            ["15.0.0.1", "10.0.0.20"],
        ),
        ([], []),
    )
    @ddt.unpack
    def test_generate_ip_range(self, ip_ranges, expected):
        ips = driver_util.generate_ip_range(ip_ranges)
        assert ips == expected

    def test_generate_ip_range_edge_cases(self):
        # Test edge cases for generate_ip_range function
        self.assertEqual(driver_util.generate_ip_range([]), [])
        self.assertEqual(driver_util.generate_ip_range(
            [["15.0.0.1", "15.0.0.1"]]), ["15.0.0.1"]
        )

    def test_generate_ip_range_large_range(self):
        # Test with a large range of IPs
        start_ip = "192.168.0.1"
        end_ip = "192.168.255.255"
        ips = driver_util.generate_ip_range([[start_ip, end_ip]])
        self.assertEqual(len(ips), 65535)


class MockClass1:
    def method1(self):
        return 1

    def _private_method(self):
        return 2


class TestDecorateMethodsWith(test.TestCase):

    def test_decorate_methods_with(self):
        decorated_cls = driver_util.decorate_methods_with(
            mock.Mock())(MockClass1)
        self.assertTrue(hasattr(decorated_cls, 'method1'))
        self.assertTrue(hasattr(decorated_cls, '_private_method'))


class MockClass2:
    @driver_util.verbose_driver_trace
    def method1(self):
        return 1


class TestVerboseDriverTrace(test.TestCase):

    def test_verbose_driver_trace_debug_true(self):
        mock_instance = MockClass2()
        with mock.patch.object(
                driver_util.LOG, 'debug') as mock_debug:
            mock_instance.method1()
            self.assertEqual(mock_debug.call_count, 2)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import copy

from oslo_config import cfg

from manila.api.validation import parameter_types
from manila.api.validation import response_types

CONF = cfg.CONF


show_request_query = {
    'type': 'object',
    'properties': {},
    'required': [],
    # TODO(jonathan): Exclude additional query string parameters in a future
    # microversion
    'additionalProperties': True,
}

index_request_query = {
    'type': 'object',
    'properties': {
        'limit': parameter_types.single_param(
            parameter_types.non_negative_integer
        ),
        # NOTE(stephenfin): This is parsed by 'common.get_pagination_params'
        # but we ignore it. We may wish to uncomment this when that is no
        # longer the case
        # 'marker': parameter_types.multi_params({
        #     'type': ['string'],
        # }),
        'offset': parameter_types.single_param(
            parameter_types.non_negative_integer
        ),
        'sort_key': parameter_types.single_param({
            'type': 'string',
            'default': 'created_at',
            # TODO(stephenfin): These are the allowed (a.k.a. legal) filter
            # keys, but we currently ignore invalid keys. We should add this in
            # a future microversion.
            # 'enum': [
            #     'id',
            #     'project_id',
            #     'request_id',
            #     'resource_type',
            #     'action_id',
            #     'detail_id',
            #     'resource_id',
            #     'message_level',
            #     'expires_at',
            #     'created_at',
            # ],
        }),
        'sort_dir': parameter_types.single_param({
            'type': 'string',
            'default': 'desc',
            # TODO(stephenfin): This should be an enum, but we currently treat
            # anything != 'desc' as 'asc'. We should make this stricter in a
            # future microversion.
            # 'enum': ['asc', 'desc'],
        }),
        'action_id': parameter_types.single_param({
            'type': 'string',
        }),
        'detail_id': parameter_types.single_param({
            'type': 'string',
        }),
        # TODO(jonathan) add enum when more message level the 'ERROR'
        'message_level': parameter_types.single_param({
            'type': 'string',
        }),
        'request_id': parameter_types.single_param({
            'type': 'string',
        }),
        'resource_id': parameter_types.single_param({
            'type': 'string',
        }),
        'resource_type': parameter_types.multi_params({
            'type': 'string',
        }),
    },
    'required': [],
    # TODO(jonathan): Exclude additional query string parameters in a future
    # microversion
    'additionalProperties': True,
}

index_request_query_v252 = copy.deepcopy(index_request_query)
index_request_query_v252['properties'].update({
    'created_since': parameter_types.single_param({
        'type': 'string',
        'format': 'date-time',
    }),
    'created_before': parameter_types.single_param({
        'type': 'string',
        'format': 'date-time',
    }),
})


_messages_response = {
    'type': 'object',
    'properties': {
        'action_id': {'type': 'string'},
        'created_at': {'type': 'string', 'format': 'date-time'},
        'detail_id': {'type': 'string'},
        'expires_at': {'type': 'string', 'format': 'date-time'},
        'id': {'type': 'string', 'format': 'uuid'},
        'links': response_types.links,
        'message_level': {
            'type': 'string',
            'enum': ['ERROR'],
        },
        'project_id': {'type': 'string'},
        'request_id': {'type': 'string'},
        'resource_id': {'type': 'string', 'format': 'uuid'},
        'resource_type': {'type': 'string'},
        'user_message': {'type': 'string'},
    },
    'required': [
        'action_id',
        'created_at',
        'detail_id',
        'expires_at',
        'id',
        'links',
        'message_level',
        'project_id',
        'request_id',
        'resource_id',
        'resource_type',
        'user_message',
    ],
    'additionalProperties': False,
}

index_response_body = {
    'type': 'object',
    'properties': {
        'messages': {
            'type': 'array',
            'items': _messages_response,
        },
        'messages_links': response_types.collection_links,
    },
    'required': ['messages'],
    'additionalProperties': False,
}

show_response_body = {
    'type': 'object',
    'properties': {
        'message': _messages_response,
    },
    'required': ['message'],
    'additionalProperties': False,
}

delete_response_body = {
    'type': 'null',
}
